#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions support for the Token and Tokens classes.
"""

from typing import List, Tuple
import warnings

class BoundaryWarning(Warning):
    pass

def _startstop(obj, start:int, stop:int) -> Tuple[int, int]:
    """
    Tool function to catch the edge of the constructed string in the following methods.
    Works for both the Token and Tokens classes, if obj is passed as self.
    Recast start and stop inside the Token.string coordinates in case obj is a Token.string.
    
    Raise a BoundaryWarning in case at least one of the the start or stop has been changed.
    """
    start_, stop_ = start, stop
    if not stop and stop!=0: stop = len(obj)
    if stop < 0 or stop > len(obj): stop = len(obj)
    if start > stop: start = stop
    if start < 0: start = 0
    if start_!=start or stop_!=stop:
        mess = "At least one of the boundaries has been modified"
        warnings.warn(mess, category=BoundaryWarning)
    return start, stop

def _isSpan(span) -> bool:
    """Returns True if `span` has all the attributes of a Span instance. False if not."""
    bools = [hasattr(span,attr) for attr in ['string','ranges','subtoksep']]
    return all(bools)

def _checkSpan(span):
    """Raises a ValueError in case `span` is not a proper Span instance"""
    if not _isSpan(span):
        raise ValueError("Accepts only Span instance")
    return None

def _isToken(token):
    """Returns True if `token` has all the attributes of a Token instance. False if not."""
    bools = [hasattr(token,attr)
             for attr in ['string','ranges','_extra_attributes',
                          'subtoksep','carry_attributes']]
    return all(bools)

def _checkToken(token):
    """Raises a ValueError in case `token` is not a proper Token instance"""
    if not _isToken:
        raise ValueError("Accepts only Token instance")
    return None

def _isTokens(tokens):
    """Returns True if all elements of `tokens` is a Token instance according to `_isToken`, otherwise returns False"""
    # to handle a non iterable
    if not any([hasattr(tokens,'tokens'),isinstance(tokens,(list,tuple))]):
        bools = [False,]
    else:
        bools = [_isToken(tok) for tok in tokens]
    return all(bools)

def _checkTokens(tokens):
    """Raises a ValueError in case `tokens` is not a proper Tokens instance."""
    if not _isTokens(tokens):
        raise ValueError("Accepts only Tokens instance")
    return None

def _checkRange(r):
    """Raises a ValueError in case r is not a compatible range object."""
    if not isinstance(r, range):
        raise ValueError("r is not an instance of Range")
    if r.start > r.stop:
        raise ValueError("r must be compatible : start<=stop")
    if r.step != 1:
        raise ValueError("r must have a step of 1")
    return None

def _checkRanges(ranges):
    """Raises a ValueError in case one range in `ranges` is not suitable for the `Span` or `Token` classes."""
    for r in ranges:
        _checkRange(r)
    return None

def _checkSameString(obj1, obj2):
    """Raises a TypeError in case the two objects have not the same .string attribute. raises an AttributeError in case at least one of the objects has no attribute .string."""
    if str(obj1.string) != str(obj2.string):
        s = obj1.__class__.__name__
        mess = "can only compare {} objects having same {}.string".format(s,s)
        raise TypeError(mess)
    return None

def _areOverlapping(r1: range, r2: range) -> bool:
    """Take two range objects, and return True if they are disjoint, otherwise return False if they overlap on some range"""
    min_ = max(r1.start,r2.start)
    max_ = min(r1.stop,r2.stop)
    return bool(max(max_-min_,0))

def _isInside(r1: range, r2: range) -> bool:
    """Take two range objects r1 and r2, and return True if r2 is in r1, otherwise return False."""
    return r1.start<=r2.start and r1.stop>=r2.stop

def _combineRanges(ranges: List[range]) -> List[range]:
    """
Take a list of range objects, and transform it such that overlapping ranges and consecutive ranges are combined. 

Exemple:
```python
_combineRanges([(12,25),(35,40)]) # -> [(12,25),(35,40),]
_combineRanges([(12,25),(26,40)]) # -> [(12,25),(26,40),]
_combineRanges([(12,26),(26,40)]) # -> [(12,40),]
_combineRanges([(12,25),(15,40)]) # -> [(12,40),]
_combineRanges([(12,25),(15,16)]) # -> [(12,25),]
```
Where all range objects have been transformed in tuples `(12,25)==range(12,25)` for illustration purpose.
The overlapping information is lost in the process.

`ranges` is a list of `range` object, all with `range.step==1` (not verified by this function, but required for the algorithm to work properly)..
    """
    if len(ranges) < 2:
        return ranges
    sorted_ranges = sorted(ranges, key=lambda x: x.start)
    combined_ranges = [sorted_ranges[0],]
    for next_range in sorted_ranges[1:]:
        previous_range = combined_ranges[-1]
        # overlapping ranges
        if previous_range.stop >= next_range.start:
            # avoid next_range inside previous_range
            if next_range.stop >= previous_range.stop:
                # extend the previous_range to next_range
                combined_ranges[-1] = range(previous_range.start,
                                            next_range.stop)
        # remain only distinct ranges since ranges are sorted
        else:
            # overwrite the next previous_range
            combined_ranges.append(next_range)
    return combined_ranges

def _findCut(ranges: List[range], cut: int, step: int=0) -> Tuple[int, int, bool]:
    """
Find the index i_ and absolute position cut_ of the cuting of ranges at relative position cut. Handle the case where the cut is inside some separator of size given by the step parameter, in which case a flag_ is raised.

Returns a tuple `(i_,cut_,flag_)` : 
    - `i_` index of the ranges element onto which the cut applies
    - `cut_` absolute position where the cut applies in the coordinates of the ranges
    - `flag_` True if the cut appears inside the separator

Raises an IndexError in case the cut is outside the range of the ranges.

Examples:
```python
ranges = [range(10,20), range(30,40)]
_findCut(ranges,5,step=0)   # -> (0, 15, False)
_findCut(ranges,10,step=0)  # -> (1, 30, False)
_findCut(ranges,10,step=1)  # -> (0, 20, False)
_findCut(ranges,11,step=1)  # -> (1, 30, False)
_findCut(ranges,10,step=2)  # -> (0, 20, False)
_findCut(ranges,11,step=2)  # -> (0, 20, True)
_findCut(ranges,12,step=2)  # -> (1, 30, False)
_findCut(ranges,20,step=1)  # -> (1, 39, False)
_findCut(ranges,21,step=1)  # -> (1, 40, False)
_findCut(ranges,22,step=1)  # -> IndexError
```

This function is the main tool to cut ranges in sub-ranges, see `_cutRanges` below.

Warning, this function does not sort the ranges, but they must be sorted for the algorithm to work properly.
    """
    cursor, i_, cut_, flag_ = 0, 0, None, False
    for i,r in enumerate(ranges):
        temp = range(cursor,cursor+step+len(r))
        if cut in temp: 
            cut_, i_ = r.start+cut-cursor, i
            if cut_ > r.stop: # cut in the separator
                cut_, flag_ = r.stop, True
            break
        cursor += len(r)+step
    if cut_ is None:
        raise IndexError("cut not found in ranges")
    return i_, cut_, flag_

def _cutRanges(ranges: List[range], cut: int, step: int=0) -> Tuple[List[range], List[range]]:
    """
Cut the ranges (given in absolute positions) at the cut (given in relative position). Handle the case where the cut is inside some separator of size given by the step parameter, in which case the entire separator length is on the left list of range objects. 

Returns two lists of range objects. The first list will correspond to the left Token (before the cut), and the second one to the right Token (after the cut).

Warnings:
    - some of the returned ranges might be empty. This is the case when the cut appears at a boundary of the range. It allows the Token instance to handle the size of the `subtoksep` separator in a smart way. This is the reason for the introduction of the `flag_` parameter in the `_findCut` function.
    - the `ranges` must be sorted for this function to work properly.

Examples:
```python
ranges = [range(10,20), range(30,40)]
_cutRanges(ranges,5,step=0)   # -> ([range(10, 15)], [range(15, 20), range(30, 40)])
_cutRanges(ranges,10,step=0)  # -> ([range(10, 20), range(30, 30)], [range(30, 40)])
_cutRanges(ranges,11,step=0)  # -> ([range(10, 20), range(30, 31)], [range(31, 40)])

# note the position of the empty range in the two exemples to come
_cutRanges(ranges,10,step=1)  # -> ([range(10, 20)], [range(20, 20), range(30, 40)])
_cutRanges(ranges,11,step=1)  # -> ([range(10, 20), range(30, 30)], [range(30, 40)])

_cutRanges(ranges,10,step=2)  # -> ([range(10, 20)], [range(20, 20), range(30, 40)])
# flag_ = True, separator is on the left : 
_cutRanges(ranges,11,step=2)  # -> ([range(10, 20), range(20, 20)], [range(30, 40)])
_cutRanges(ranges,12,step=2)  # -> ([range(10, 20), range(30, 30)], [range(30, 40)])

_cutRanges(ranges,20,step=1)  # -> ([range(10, 20), range(30, 39)], [range(39, 40)])
_cutRanges(ranges,21,step=1)  # -> ([range(10, 20), range(30, 40)], [range(40, 40)])
_cutRanges(ranges,22,step=1)  # -> IndexError
```
    """
    i_, cut_, flag_ = _findCut(ranges,cut,step)
    if not flag_:
        r1 = ranges[:i_] + [range(ranges[i_].start,cut_),]
        r2 = [range(cut_,ranges[i_].stop),] + ranges[i_+1:]
    else: # cut in the separator: the entire separator is on the left
          # range(cut_,cut_) is required to give the correct length of the string
          # in Token instance, which knows the size of subtoksep
        r1 = ranges[:i_] + [range(ranges[i_].start,cut_),range(cut_,cut_),]
        r2 = ranges[i_+1:]
    return r1,r2

def _removeRange(ranges: List[range], r: range) -> List[range]:
    """
Remove the range r from the list of ranges.

Examples:
```python
ranges = [range(10,20), range(30,40)]
_removeRange(ranges,range(0,10))   # -> [range(10, 20), range(30, 40)]
_removeRange(ranges,range(0,11))  # -> [range(11, 20), range(30, 40)]
_removeRange(ranges,range(11,15))  # -> [range(10, 11), range(15, 20), range(30, 40)]

_removeRange(ranges,range(25,30))  # -> [range(10, 20), range(30, 40)]
_removeRange(ranges,range(0,25))  # -> [range(30, 40)]

_removeRange(ranges,range(11,11))  # ->[range(10, 20), range(30, 40)]

_removeRange(ranges,range(30,31))  # -> [range(10, 20), range(31, 40)]
_removeRange(ranges,range(100))  # -> []

_removeRange(ranges,range(15,35))  # -> [range(10, 15), range(35, 40)]
```
    """
    new_ranges = []
    for r_ in ranges:
        r_temp = [r_,]
        if r.start in r_ and r.stop in r_:
            if r_.start==r.start:
                r_temp = [range(r.stop,r_.stop),]
                # if r.stop==r_.stop, one has r.stop not in r_
            else:
                r_temp = [range(r_.start,r.start),range(r.stop,r_.stop),]
        elif r.start in r_:
            r_temp = [range(r_.start,r.start),]
        elif r.stop in r_:
            r_temp = [range(r.stop,r_.stop),]
        elif r.start < r_.start:
            if r.stop > r_.stop:
                r_temp = []
            elif r.stop in r_:
                r_temp = [range(r_.start,r.stop),]
        new_ranges += r_temp
    return [r for r in _combineRanges(new_ranges) if r]

def _fusionAttributesList(attributesList):
    """Take a list of dictionnaries, and return a dictionnary of lists"""
    attributesKeys_ = set(k for d in attributesList for k in d.keys())
    attributesDict_ = {k:list() for k in attributesKeys_}
    for attributes in attributesList:
        for attr in attributesKeys_.intersection(attributes.keys()):
            attributesDict_[attr].append(attributes[attr])
        for attr in attributesKeys_.difference(attributes.keys()):
            attributesDict_[attr].append({})
    return attributesDict_

def _fusionAttributesDict(attributesDict):
    """Take a dictionnaries of lists, and return a dictionnary of dictionnaries of lists"""
    attributesDictDict_ = {}
    for key,attributeslist in attributesDict.items():
        attributesDictDict_[key] = _fusionAttributesList(attributeslist)
    return attributesDictDict_

def _fusionAttributes(attributes):
    """Apply the two above methods in a raw"""
    attributesDict_ = _fusionAttributesList(attributes)
    return _fusionAttributesDict(attributesDict_)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Merge sort algorithm, for a given order relation.

Adpated from https://www.geeksforgeeks.org/merge-sort/
"""

def sort(array, relation='__lt__'):
    
    if len(array) > 1:
         # Finding the mid of the array
        mid = len(array)//2
        # Dividing the array elements
        left_array = array[:mid]
        # into 2 halves
        right_array = array[mid:]
        # Sorting the first half
        sort(left_array, relation=relation)
        # Sorting the second half
        sort(right_array, relation=relation)
        i = j = k = 0
        # Copy data to temp arrays L[] and R[]
        while i < len(left_array) and j < len(right_array):
            if getattr(left_array[i], relation)(right_array[j]):
                array[k] = left_array[i]
                i += 1
            else:
                array[k] = right_array[j]
                j += 1
            k += 1
  
        # Checking if any element was left
        while i < len(left_array):
            array[k] = left_array[i]
            i += 1
            k += 1
  
        while j < len(right_array):
            array[k] = right_array[j]
            j += 1
            k += 1


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from pkg_resources import get_distribution
__version__ = get_distribution("tokenspan").version

from .tokentokens import Token, Tokens
from .span import Span





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
`Token` and `Tokens` class

`Token` sub-classes the `string` class in Python, thus enabling basic usages of
string (as e.g. split, isupper, lower, ... see 
https://docs.python.org/3.8/library/string.html)
in addition to enabling additional attribute in the flow, for later 
compatibility with other packages. Its main original methods are
 - `partition_token(start,stop)` : which partitions the initial `Token` in three
 new `Token` instance, collected in a unique `Tokens` instance (see below)
 - `split_token([(start,end),(start2,end2), ...])` : which splits the `Token`
 in several instances grouped in a single `Tokens` object
 - `slice(step)` : which slices the initial string in overlapping sub-strings,
 all grouped in a single `Tokens` instance
 
 `Tokens` collects the different `Token` instances in order to let the initial 
`Token` instance still glued somehow. It also allows to come back to `Token`
instances using its original methods : 
     - `undo(start,stop,step)` : which generate a unique `Token` from the `Tokens`
     elements `Tokens[start:stop:step]`
     - `slice(start,stop,step)` : which slices the list of `Token` instance and 
     glue them in some overlapping strings. This is still a `Tokens` instance

"""

from .span import Span
from .tools import _checkRange, _checkRanges, _startstop, _removeRange
from .tools import _checkToken, _isToken, _checkTokens, _checkSameString
from .tools import _combineRanges, _fusionAttributes, _cutRanges

_string_methods = ['upper','lower','swapcase','capitalize','casefold',
                    'isalnum','isalpha','isascii','isdecimal','isdigit',
                    'isidentifier','islower','isnumeric','isprintable',
                    'isspace','istitle','isupper']

_string_methods_with_args = ['startswith','endswith','encode']

_token_methods = ['get_subToken','attributes','keys','values','items',
                  'fusion_attributes','append','remove',
                  'append_range','remove_range','start','stop',
                  '_append_range','_remove_range',
                  'union','difference','intersection','symmetric_difference',
                  'copy','set_string_methods',
                  'slice','split','partition','_prepareTokens',]

class Token(Span):
    """
Subclass of the Python string class.
It allows manipulating a string as a usual Python object, excepts it 
returns Token instances. Especially for Token instances are the 
methods : 
 - `split(s)`, which splits the string everytime the string 's' appears
 - `isupper()` (for instance), which tells whether the string is 
 uppercase or not
 - `lower()` or `upper()`, to make the `Token` lower-/upper-case.
See more string methods on [the Python standard library documentation](https://docs.python.org/3/library/stdtypes.html#text-sequence-type-str).

In addition, having class associated to a Token allows to add custom
   	attribute and method at any moment during its use.

    """ 
    
    def __init__(self,
                 string='',
                 ranges=None,
                 subtoksep=chr(32),
                 carry_attributes=True):
        """
`Token` object is basically a `string` with a `ranges` (list of range) position. Its basic usage is : 
 - `string` extracted from all intervals defined in the `ranges` list
and its attributes are
 - `Token.string`	 -> a string
 - `Token.ranges`	 -> a list of ranges
 - `Token.subtoksep` -> a string, preferably of length 1
         """
        super().__init__(string=string,ranges=ranges,subtoksep=subtoksep)
        self.carry_attributes = bool(carry_attributes)
        self._extra_attributes = set(_token_methods)
        return None
    
    def __repr__(self):
        """Return the two main arguments (namely the `string` and the number of ranges) of a `Token` instance in a readable way."""
        mess = "Token('{}', ".format(str(self))
        mess += '['+','.join('('+str(r.start)+','+str(r.stop)+')' 
                             for r in self.ranges)
        mess += "])"
        return mess   
    
    def __getattr__(self,name):
        """Add some string methods to the Token objects."""
        string_methods = _string_methods + _string_methods_with_args
        if name in string_methods: # only handle str methods here
            def method(string, *args, **kwargs):
                value = getattr(string, name)(*args, **kwargs)
                return value
            res = method.__get__(str(self))
            return res
        m = "Token object has no attribute {}".format(name)
        raise AttributeError(m)

    def setattr(self,name,dic=dict(),**kwargs):
        """
Add attribute to the `Token` instance. Can be copied to a new instance using `Token.copy()`.

Call this function as `setattr('string_name',dict(a=1,b=2),c=3,d=4)` where all arguments, except the first one, are optionnal. One can pass the attributes dictionnaries either as usual dictionnaries : `dict(a=1,b=2)` or `{'a':1, 'b':2}` or directly as keywor arguments : `c=3,d=4`, or both methods, in which case all the values in the different dictionanries will be concatenated. The first argument must be a string, which will serve to append the corresponding attribute to the Token instance, e.g. `Token.string_name` will exist in the above example.

Raise an `AttributeError` in case the attribute already exists. It is still possible to update the attribute (i.e. it is not protected) in the usual way : `Token.string_name.update(a=0,b=1)` will overide `a` and `b` keys in `Token.string_name` dictionnary.
        """
        n = str(name)
        if n in self._extra_attributes:
            m = "Attribute {} already exist".format(n)
            raise AttributeError(m)
        if dic:
            kwargs = {**dic,**kwargs}
        setattr(self,str(name),dict(**kwargs))
        self._extra_attributes.update((n,))
        return self
    
    def copy(self,reset_attributes=False):
        """Returns a copy of the actual `Token` instance. Take care of the `Token.attributes` if they are created using `Token.setattr` and if `reset_attributes` is `False` (value per default)."""
        tok = Token(string=self.string,
                    ranges=self.ranges,
                    subtoksep=self.subtoksep,
                    carry_attributes=self.carry_attributes)
        if not reset_attributes and self.carry_attributes:
            for name in self.attributes:
                kwargs = getattr(self,name)
                tok.setattr(name,**kwargs)
        return tok
    
    def set_string_methods(self,):
        """Add the basic Python string methods to the attribute `string_methods` of the Token."""
        string_methods = _string_methods
        method_dict = {name: getattr(str(self),name)()
                       for name in string_methods}
        self.setattr('string_methods',**method_dict)
        return self

    def __eq__(self,token):
        """
Verify whether the actual instance of Token and an extra ones have the same attributes. 
        
Raise an Exception in case the attributes are the same, but the extra_attributes are not all equals, despite the two instances have the same names of extra_attributes.

Raise a ValueError when one object is not a Token instance
        """
        _checkToken(token)
        bools = [token.string==self.string,
                 token.ranges==self.ranges,
                 token.attributes==self.attributes,
                 token.subtoksep==self.subtoksep,]
        if all([self.carry_attributes,token.carry_attributes]+bools):
            bools.extend([getattr(token,attr)==getattr(self,attr)
                          for attr in token.attributes])
            bools.extend([getattr(token,attr)==getattr(self,attr)
                          for attr in self.attributes])
        return all(bools)
    
    @property
    def attributes(self,):
        """Returns the name of the attributes in a frozenset"""
        return frozenset(self._extra_attributes.difference(_token_methods))
    def keys(self,):
        """Returns the keys of the attributes in a generator"""
        return (att for att in self.attributes)
    def values(self,):
        """Returns the values of the attributes in a generator"""
        return (getattr(self,att) for att in self.attributes)
    def items(self,):
        """Returns the tuple (key,value) of the attributes in a generator"""
        return ((att,getattr(self,att)) for att in self.attributes)

    def get_subToken(self,n):
        """
Get the Token associated to the ranges elements n (being an integer or a slice).
Return a Token.
Raise a IndexError in case n is larger than the number of ranges in self.ranges.
        """
        r = self.ranges[n]    
        token = self.copy()
        token.ranges = [r,]            
        return token

    @property
    def subTokens(self,):
        """
Get the Token associated to each Token.ranges in a list
Return a list of Token. Keep the attributes in case Token.carry_attributes is True.
        """
        return Tokens([self.get_subToken(i) for i in range(len(self.ranges))])
    
    def fusion_attributes(self,token):
        """Fuse the attributes of the present instance with the `token` one."""
        _checkToken(token)
        if not all([self.carry_attributes,token.carry_attributes]):
            mess = "Fusion aborted because at least one carry_attributes is False"
            raise AttributeError(mess)
        self_attrs = {k:v for k,v in self.items()}
        token_attrs = {k:v for k,v in token.items()}
        new_attributes = _fusionAttributes([self_attrs,token_attrs])
        new_token = self.copy(reset_attributes=True)
        for attr,kwargs in new_attributes.items():
            new_token.setattr(attr,**kwargs)
        return new_token

    def union(self, token):
        """
Quasi-Alias for append. 
Returns a new Token instance with Token.ranges = self.append(r).ranges
For compatibility with set terminology.
        """
        _checkToken(token)
        _checkSameString(self,token)
        new_token = self.copy()
        new_token.append_ranges(token.ranges)
        if all([self.carry_attributes,token.carry_attributes]):
            new_token = new_token.fusion_attributes(token)
        else:
            new_token = new_token.copy(reset_attributes=True)        
        return new_token 

    def difference(self,token):
        """
Quasi-Alias for remove.
Return a new Token instance with Token.ranges = self.remove(r).ranges
For compatibility with set terminology.
        """
        _checkToken(token)
        _checkSameString(self,token)
        new_token = self.copy()
        new_token.remove_ranges(token.ranges)
        flag = True if not new_token.ranges else False
        if all([self.carry_attributes,token.carry_attributes]):
            new_token = new_token.fusion_attributes(token)
        else:
            new_token = new_token.copy(reset_attributes=True)
        if flag:
            new_token.ranges = []
        return new_token
    
    def __AmB_BmA(self,token):
        """Utility to calculate the intersection and symmetric_difference"""
        AmB, BmA = self.copy(), token.copy()
        AmB.remove_ranges(token.ranges)
        BmA.remove_ranges(self.ranges)
        return AmB, BmA
    
    def intersection(self,token):
        """
Return a new Token whose Token.ranges is the intersecting ranges with the given ranges in the self instance.
        """
        _checkToken(token)
        _checkSameString(self,token)
        # one uses the fact that AxB = A+B-(A-B)-(B-A)
        newToken = self.copy()
        AmB, BmA = self.__AmB_BmA(token)
        newToken.append_ranges(token.ranges)
        newToken.remove_ranges(BmA.ranges)
        newToken.remove_ranges(AmB.ranges)
        ranges = [r for r in _combineRanges(newToken.ranges) if r]
        if not ranges:
            newToken = Token()
            newToken.ranges = ranges
        else:
            newToken.ranges = ranges
        return newToken
    
    def symmetric_difference(self,token):
        """
Return a new Token instance whose Token.ranges is the symmetric difference ranges with the given ranges in the self instance
        """
        _checkToken(token)
        _checkSameString(self,token)
        # one uses the fact that A/B = (A-B)+(B-A)
        newToken = self.copy()
        AmB, BmA = self.__AmB_BmA(token)
        AmB.append_ranges(BmA.ranges)
        newToken.ranges = [r for r in _combineRanges(AmB.ranges) if r]
        return newToken
      
    def _prepareTokens(self,ranges,reset_attributes):
        """remove empty ranges and keep attributes and construct the Tokens object"""
        tokens = Tokens([Token(string=self.string,
                               ranges=r,
                               subtoksep=self.subtoksep,
                               carry_attributes=self.carry_attributes)
                         for r in ranges])
        if not reset_attributes and self.carry_attributes:
            toks = [tok.fusion_attributes(self) for tok in tokens]
            tokens = Tokens(toks)
        return tokens
    
    def partition(self,start,end,remove_empty=False,reset_attributes=False):
        """
Split the `Token.string` in three `Token` objects : 
 - `string[:start]`
 - `string[start:stop]`
 - `string[stop:]`
and put all non-empty `Token` objects in a `Tokens` instance.

It acts a bit like the `str.partition(s)` method of the Python
`string` object, but `partition_token` takes `start` and `end` 
argument instead of a string. So in case one wants to split a string in three 
sub-strings using a string 's', use `Token.partition(s)` instead, 
inherited from `str.partition(s)`.

NB : `Token.partition(s)` has no `non_empty` option.

| Parameters | Type | Details |
| --- | --- | --- | 
| `start` | int | Starting position of the splitting sequence. |
| `end` | int | Ending position of the splitting sequence. |
| `non_empty` | bool. Default is `False` | If `True`, returns a `Tokens` instance with only non-empty `Token` objects. see __bool__() method for non-empty `Token` |

| Returns | Type | Details |
| --- | --- | --- | 
| tokens | `Tokens` object | The `Tokens` object containing the different non-empty `Token` objects. |
        """
        spans = super().partition(start,end,remove_empty=remove_empty)
        ranges = [s.ranges for s in spans]
        tokens = self._prepareTokens(ranges,reset_attributes)
        return tokens
    
    def split(self,cuts,remove_empty=False,reset_attributes=False):
        """
Split a text as many times as there are entities in the cuts list.
Return a `Tokens` instance.

This is a bit like `str.split(s)` method from Python `string`
object, except one has to feed `Token.split_token` with a full list
of `(start,end)` tuples instead of the string 's' in `str.split(s)`
If the `(start,end)` tuples in cuts are given by a regex re.finditer
search, the two methods give the same thing. So in case one wants
to split a string in several `Token` instances according to a 
string 's' splitting procedure, use `Token.split(s)` instead of 
`Token.split_token([(start,end), ...])`.

| Parameters | Type | Details |
| --- | --- | --- | 
| `cuts` | a list of `(start,end,)` tuples. start/end are integer | Basic usage is to take these cuts from [`re.finditer`](https://docs.python.org/3/library/re.html#re.finditer). |

| Return | Type | Details |
| --- | --- | --- | 
| `tokens` | A `Tokens` object | A `Tokens` instance containing all the `Token` instances of the individual tokens. |

        """
        _checkRanges(cuts)
        spans = super().split(cuts, remove_empty=remove_empty)
        ranges = [s.ranges for s in spans]
        tokens = self._prepareTokens(ranges,reset_attributes)      
        return tokens

    def slice(self,start=0,stop=None,size=1,step=1,
              remove_empty=False,reset_attributes=False):
        """
Cut the `Token.string` in overlapping sequences of strings of size `size` by `step`,
put all these sequences in separated `Token` objects, and finally 
put all theses objects in a `Tokens` instance.

| Parameters | Type | Details |
| --- | --- | --- | 
| `size` | int | The size of the string in each subsequent Token objects. |
| `step` | int | The number of characters skipped from one Token object to the next one. |

| Returns | Type | Details |
| --- | --- | --- | 
| `tokens` | `Tokens` object | The `Tokens` object containing the different `Token` sliced objects. |
        """
        spans = super().slice(start=start,
                              stop=stop,
                              size=size,
                              step=step,
                              remove_empty=remove_empty)
        ranges = [s.ranges for s in spans]
        tokens = self._prepareTokens(ranges,reset_attributes)
        return tokens



class Tokens():
    """
A tool class for later uses in `Tokenizer` class (not documented in this module).
This is mainly a list of `Token` objects, with additional methods to 
implement string manipulations, and go back to individual `Token` instances.
    """
    def __init__(self,tokens=list()):
        """
`Tokens` instance is just a list of `Token` instances, called

 - `Tokens.tokens`	-> a list attribute.
 
The only verification is that the `Token.copy()` methods works.
        """
        _checkTokens(tokens)
        self.tokens = list(tok.copy() for tok in tokens.copy())
        return None
    
    def __repr__(self,):
        """
Representation of the `Tokens` class, printing the number of `Token`
instances inside the `Tokens` one, and the concatenated string from
all `Tokens` instances.
        """
        mess = "Tokens({} Token)".format(len(self))
        m = len(mess)+2
        mess += " : \n"+"-"*m+"\n"
        for token in self:
            mess += "{}\n".format(token)
        return mess
    
    def __len__(self,):
        """
Return the number of `Token` instances in the `Tokens` object.
        """
        return len(self.tokens)
    
    def __add__(self,tokens):
        """
Add two `Tokens` instances in the same way `list` can be concatenated :  
by concatenation of their tokens list of `Token` instances.
Returns a new Tokens instance.
        """
        _checkTokens(tokens)
        toks = list(tok.copy() for tok in self.tokens)
        toks += list(tok.copy() for tok in tokens)
        return Tokens(tokens=toks)
    
    def __str__(self,):
        """
Return the concatenated string of all the `Token` instances in the
`Tokens.tokens` attribute. Each Token string is separated from its neighborhood by the `_NEW_Token_` string.
        """
        strings = [str(token) for token in self.tokens]
        string = '_NEW_Token_'.join(strings)
        return string
    
    def __getitem__(self,n):
        """
Return either a `Tokens` new instance in case a slice is given,
or the `Token` instance correspondig to the position n in case
an integer is catched as argument.
        """
        if isinstance(n,slice):
            return Tokens(self.tokens[n])
        if type(n) == int:
            return self.tokens[n]
        if type(n) == str:
            return self.attributes_values[n]
        return AttributeError('argument must be slice or integer')
    
    def __eq__(self,tokens):
        """Return True if all elements of self are the same as in tokens."""
        self._checkTokens(tokens)
        return all(t1==t2 for t1,t2 in zip(self,tokens))
    
    def __contains__(self,token):
        """Return True if the Token is included in one of the Token in self"""
        _checkToken(token)
        return any(token in tok for tok in self.tokens)
    
    def __bool__(self,):
        """Return True if any of the Token in Tokens is True, otherwise False"""
        return any(tok for tok in self.tokens)
    
    def copy(self,reset_attributes=False):
        """Make a copy of the Tokens object"""
        return Tokens([tok.copy(reset_attributes=reset_attributes) 
                       for tok in self.tokens])
    
    @property
    def attributes_keys(self,):
        """Find all the _extra_attributes in all Token objects composing the Tokens, returns a frozenset"""
        return frozenset([attr for tok in self for attr in tok.attributes])
    
    @property
    def attributes_map(self,):
        """Find all the Token indexes per _extra_attributes, returns a dictionnary map."""
        attr_map = {key:list() for key in self.attributes_keys}
        for i,token in enumerate(self):
            for attr in token.attributes:
                attr_map[attr].append(i)
        return attr_map
    
    @property
    def attributes_values(self,):
        """Find all the values of all the _extra_attributes of all the Token objects composing the Tokens object, returns a dictionnary {attribute: [list of dictionnaries of for the attribute for Token 1, then for Token 2, ...], with one entry per Token."""
        attr_values = {key:list() for key in self.attributes_keys}
        for k,v in self.attributes_map.items():
            attr_values[k] = [getattr(self[i],k) if i in v else dict()
                              for i in range(len(self))]
        return attr_values
    
    def keys(self,):
        """Returns the generator of all the attributes present in the Tokens instance, as given by each Token element of Tokens."""
        return (k for k in self.attributes_keys)
    
    def values(self,):
        """Returns the generator of all the attributes values in the Tokens instance, as given by each Token element of Tokens."""
        return (self.attributes_values[k] for k in self.attributes_keys)
    
    def items(self,):
        """Returns the generator of the tuples (keys, values), as given by these methods."""
        return ((k,self.attributes_values[k]) for k in self.attributes_keys)
    
    def has_attribute(self,attr):
        """Returns a new Tokens instance, with only those Token objects having the required attribute in parameter."""
        if attr not in self.attributes_keys:
            raise KeyError('attribute not found')
        tokens = [self[i].copy() for i in self.attributes_map[attr]]
        return tokens
    
    def append(self,token):
        """Append a `Token` instance to the actual `Tokens` instance."""
        _checkToken(token)
        self.tokens.append(token)
        return self
    
    def extend(self,tokens):
        """Extend a list of tokens to the actual Tokens instance."""
        _checkTokens(tokens)
        self.tokens.extend(tokens)
        return self
    
    def insert(self,position,token):
        """Insert a `Token` instance to the actual `Tokens` instance at position `position (an integer)`
        """
        _checkToken(token)
        self.tokens = self.tokens[:position]+[token,]+self.tokens[position:]
        return self
    
    def join(self,start=0,stop=None,step=1,
             reset_attributes=False):
        """
Glue the different `Token` objects present in the `Tokens` instance at 
position `[start:stop:step]`.

Return a `Token` instance.
 - `step = 1` : undo the `Token.split_token` or `Token.partition_token`
methods
 - `overlap step-1` : undo the `Token.slice(step)` method

| Parameters | Type | Details |
| --- | --- | --- | 
| `start` | int, optional. The default is 0. | Starting `Token` from which the gluing starts.  |
| `stop` | int, optional. The default is `None`, in which case the stop is at the end of the string. | Ending `Token` at which the gluing stops. |
| `step` | int, optional. The default is 1. | The step in the `Tokens.tokens` list. If `step = 1`, undo the `Token.split_token` method. |
|`overlap`| int, optional. The default is 0. | The number of characters that will be drop from each `Token.string` before concatenating it to the undone one. If `overlap = step-1` from  `Token.slice(step)`, undo the `Token.slice` method.|

_Remark:_ the reason why `glue(step)` does not revert the `Token.slice(step)` process is because one has

```python
token = Token(string=text)
tokens = token.slice(size)
undone = tokens.glue(step=size)
assert undone.string == text[:-(len(text)%size)]
```
for any `text` (a string) and `size` (an int). So when `len(text)%size=0` everything goes well, but when there are rest string, one has to do: 
    
```python
token = Token(string=text)
tokens = token.slice(size)
undone = tokens.glue(overlap=size-1)
assert undone.string == text

```

| Returns  | Type | Details |
| --- | --- | --- | 
| `tokens` | A `Token` instance | The container containing the glued string. |

        """
        start, stop = _startstop(self,start,stop)
        tokstemp = self.tokens[start:stop:step]
        if len(tokstemp)<1:
            return Token()
        token = tokstemp[0].copy(reset_attributes=True)
        ranges = [r for tok in tokstemp for r in tok.ranges]
        ranges = _combineRanges(ranges)
        attributes = [{k:v for k,v in tok.items()} for tok in tokstemp]
        attributes = _fusionAttributes(attributes)
        token.ranges = ranges
        for name,attr in attributes.items():
            token.setattr(name,**attr)
        return token
    
    def slice(self,start=0,stop=None,size=1,step=1):
        """
Glue the different `Token` objects present in the `Tokens.tokens` list and returns a list of `Token objects` with overlapping strings among the different `Token` objects, all together grouped in a `Tokens` instance.

| Parameters | Type | Details |
| --- | --- | --- | 
| `start` | int, optional. The default is 0. | Starting `Token` from which the gluing starts. |
| `stop` | int, optional. The default is `None`, in which case the stop is at the end of the string. | Ending `Token` at which the gluing stops. |
| `size` |int, optional. The default is 1. | Size of the span of each slice. |
| `step` | int, optional. The default is 1. | The step in the Tokens.tokens list. If `step = 1`, give back the initial `Tokens` object. If `step = n > 1`, give some n-grams `Token` by `Token`. |

| Returns | Type | Details |
| --- | --- | --- | 
| `tokens` | `Tokens` object | `Tokens` objects containing the list of n-grams `Token` objects. |

        """
        start, stop = _startstop(self,start,stop)
        tokens = Tokens()
        for i in range(start,stop-size+1,step):
            tok = self.join(i,i+size)
            tokens.append(tok)
        return tokens



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
`Span` class

Its main original methods are
 - `partition(range(start,stop))` : which partitions the initial `Span` in three
 new `Span` instance, collected in a unique `Spans` instance (see below)
 - `split([range(start,end),range(start2,end2), ...])` : which splits the `Span`
 in several instances grouped in a a list of `Span` objects
 - `slice(start,stop,size,step)` : which slices the initial string from position 
`start` to position `stop` by `step` in sub-strings of size `size`, all grouped 
in a list of Span objects

In addition, one can compare two `Span` using the non-overlapping order `<` and `>` or the overlapping-order `<=` and `>=` when two `Span`s overlap.

Finally, since a `Span` can be seen as a selection of a set of character positions from the parent string, one can apply the basic set operations to two `Span`s, in order to construct more elaborated `Span` instance. 

"""

import warnings

from hashlib import sha1
from .tools import _checkRange, _checkRanges, _startstop, _removeRange
from .tools import _checkSpan, _checkSameString
from .tools import _combineRanges, _cutRanges

class Span():
    """
`Span` object is basically a collection of a parent string named `string`, and a `ranges` collection (list of range) of positions. Its basic usage is its `str` method, which consists in a string extracted from all intervals defined in the `ranges` list, and joined by the `subtoksep` separator.
    """ 
    
    def __init__(self,
                 string='',
                 ranges=None,
                 subtoksep=chr(32),
                 encoding='utf-8'):
        """
`Span` attributes are
 - `Span.string`	 -> a string, default is empty
 - `Span.ranges`	 -> a list of ranges, default is None, in which case the ranges are calculated to contain the entire string
 - `Span.subtoksep`  -> a string, preferably of length 1, default is a white space ' '
 - `Span.encoding`   -> a string, representing the encoding, default is `'utf-8'`
        """
        if isinstance(string, Span):
            _spankeys = ['string', 'ranges', 'subtoksep', 'encoding']
            args = {k:v for k,v in string.__dict__.items() if k in _spankeys}
            self._init_span(**args)
        else:
            self._init_span(string=string,
                            ranges=ranges,
                            subtoksep=subtoksep,
                            encoding=encoding)
        return None
    
    def _init_span(self, 
                   string='', 
                   ranges=None, 
                   subtoksep=chr(32),
                   encoding='utf-8'):
        self.string = str(string)
        self.subtoksep = str(subtoksep)
        self.encoding = str(encoding)
        # construct the ranges only when it is not passed as argument
        # keep empty list in case it is a list
        if ranges is None:
            self.ranges = [range(len(self.string)),] if len(self.string) else []
            return None
        self.ranges = ranges
        # check if ranges are correct
        _checkRanges(self.ranges)
        # correct ranges in case some are over the string size
        ranges_ = [range(*_startstop(self.string,r.start,r.stop))
                   for r in self.ranges]
        # withdraw overlapping ranges
        self.ranges = _combineRanges(ranges_)
        return None

    def _append_range(self, r):
        """Utility that appends a range in self.ranges"""
        _checkRange(r)
        # recasts r.start and r.stop inside the range(0,len(self.string))
        start, stop = _startstop(self.string,r.start,r.stop)
        self.ranges.append(range(start,stop))
        return None

    def append_range(self, r):
        """
Append a range object to self.ranges. The range r must be given in absolute coordinates.
Return self (append in place).
Raise a `ValueError` in case r is not a range object.
Raise a `BoundaryWarning` in case r has start or stop attributes outside the size of Span.string, in which case thse parameters are recalculated to fit Span.string (being either 0 for start or len(Span.string) for stop).
        """
        self._append_range(r)
        self.ranges = [r for r in _combineRanges(self.ranges) if r]
        return self
    
    def append_ranges(self, ranges):
        """
Append a list of range objects to self.ranges. This method applies `append_range` several times, so please see its documentation for more details.
        """
        for r in ranges:
            self._append_range(r)
        self.ranges = [r for r in _combineRanges(self.ranges) if r]
        return self

    def _remove_range(self,r):
        """Utility that removes a range from self.ranges"""
        _checkRange(r)
        self.ranges = _removeRange(self.ranges,r)
        return None

    def remove_range(self,r):
        """
Remove the range r from Span.ranges. The range r must be given in absolute coordinates.
Return self (remove in place).
In case the range r encompass the complete string, there is no more Span.ranges associated to the outcome of this method.
        """
        self._remove_range(r)
        return self

    def remove_ranges(self,ranges):
        """
Remove a list of range objects toself.ranges. This method applies `remove_range` several times, so please see its documentation for more details.
        """
        for r in ranges:
            self._remove_range(r)
        return self

    def __len__(self):
        """Return the length of the string associated with the Span"""
        l = sum(r.stop-r.start for r in self.ranges) 
        l += (len(self.ranges)-1)*len(self.subtoksep)
        return max(l,0) # in case ranges=[], l<0
    
    def __repr__(self):
        """Return the two main elements (namely the `string` and the `ranges` attributes) of a `Span` instance in a readable way."""
        mess = '{}'.format(self.__class__.__name__)
        mess += "('{}', ".format(str(self))
        mess += '['+','.join('('+str(r.start)+','+str(r.stop)+')' 
                             for r in self.ranges)
        mess += "])"
        return mess   
    
    def __str__(self):
        """
`str(Span)` method returns the recombination of the extract of each `Span.subSpan` from the `Span.string` attribute corresponding to all its `Span.ranges` attribute, and joined by the `Span.subtoksep` character.
        """
        return self.subtoksep.join(self.string[r.start:r.stop] 
                                   for r in self.ranges)

    def __hash__(self,):
        """
Make the Span object hashable, such that it can serve for set and dict.keys. Span is constructed on the unicity of the Span object, that is, this is the hash of the string made of the parent string, plus the string representation of the instance, including subtoksep. Everything is then converted to hashlib.sha1.hexdigest
        """
        _bytes_ = bytes(repr(self) + self.string + self.subtoksep,
                        encoding=self.encoding)
        return int(sha1(_bytes_).hexdigest(), 16)
        
    
    def __contains__(self,s):
        """If the object to be compared with is a Span related to the same string as this instance, check whether the ranges are overlapping. Otherwise, check whether the string str(s) (which transforms the other Span instance in a string in case s is not related to the same string) is a sub-string of the `Span` instance."""
        try:
            _checkSpan(s)
        except ValueError:
            b = str(s) in str(self)
        else:
            if s.string==self.string:
                b = any(r1.start<=r2.start and r1.stop>=r2.stop 
                        for r1 in self.ranges for r2 in s.ranges)
            else:
                b = False
        return b
    
    def __bool__(self):
        """Return `True` if the `Span.ranges` is non-empty, otherwise return `False`"""
        return bool(len(self))
    
    def __getitem__(self,n):
        """
Allow slice and integer catch of the elements of the string of `Span`.
Return a string.

Note: As for the usual Python string, a slice with positions outside str(Span) will outcome an empty string, whereas Span[x] with x>len(Span) would results in an IndexError.
        """
        return str(self)[n]

    def get_subSpan(self, n):
        """Alias for get_subspan"""
        warnings.warn("prefer using get_subspan, this method will disapear after version 1.0",
                      DeprecationWarning)
        return self.get_subspan(n)
    
    def get_subspan(self, n):
        """
Get the Span associated to the ranges elements n (being an integer or a slice).
Return a Span object.
Raise an IndexError in case n is larger than the number of ranges in self.ranges.
        """
        span = self.__class__(string=self.string,
                              ranges=[self.ranges[n],],
                              subtoksep=self.subtoksep)          
        return span
    
    @property
    def subSpans(self,):
        """Alias for subspans"""
        warnings.warn("prefer using subspan, this method will disapear after version 1.0",
                      DeprecationWarning)
        return self.subspans
    
    @property
    def subspans(self,):
        """
Get the Span associated to each Span.ranges in a Span object.
Return a list of Span objects.  
        """
        return [self.get_subspan(i) for i in range(len(self.ranges))]

    def __eq__(self, span):
        """
Verify whether the actual instance of Span and an extra ones have the same attributes. 

Returns a boolean.

Raise a ValueError when one object is not a Span instance
        """
        _checkSpan(span)
        bools = [span.string==self.string,
                 span.ranges==self.ranges,
                 span.subtoksep==self.subtoksep,]
        return all(bools)

    def __neq__(self, span):
        """
Verify whether the actual instance of Span and an extra have any differing attribute. 

Returns a boolean.

Raise a ValueError when one object is not a Span instance
        """
        _checkSpan(span)
        bools = [span.string!=self.string,
                 span.ranges!=self.ranges,
                 span.subtoksep!=self.subtoksep,]
        return any(bools)

    def __add__(self, span):
        """If the two Span objects have same strings, returns a new Span object with combined ranges of the initial ones."""
        return self.union(span)

    def __sub__(self, span):
        """If the two Span objects have same strings, returns a new Span object with ranges of self with Span ranges removed. Might returns an empty Span."""
        return self.difference(span)

    def __mul__(self, span):
        """If the two Span objects have same strings, returns a new Span object with ranges of self having intersection with Span ranges removed. Might returns an empty Span."""
        return self.intersection(span)

    def __truediv__(self, span):
        """If the two Span objects have same strings, returns a new Span object with ranges of self having symmetric_difference with Span ranges removed. Might returns an empty Span."""
        return self.symmetric_difference(span)
    
    @property
    def start(self,):
        """Returns the starting position (an integer) of the first ranges. Make sense only for contiguous Span."""
        return self.ranges[0].start
    @property
    def stop(self,):
        """Returns the ending position (an integer) of the last ranges. Make sense only for contiguous Span."""
        return self.ranges[-1].stop
    
    def __lt__(self, span):
        """Returns True if Span is entirely on the left of span (the Span object given as parameter). Make sense only for contiguous Span and non-empty ones."""
        _checkSpan(span)
        _checkSameString(self, span)
        return self.stop <= span.start
    
    def __gt__(self, span):
        """Returns True if Span is entirely on the right of span (the Span object given as parameter). Make sense only for contiguous Span and non-empty ones."""
        _checkSpan(span)
        _checkSameString(self, span)
        return span.stop <= self.start
    
    def __le__(self, span):
        """Returns True if Span is partly on the left of span (the Span object given as parameter). Make sense only for contiguous Span and non-empty ones."""
        _checkSpan(span)
        _checkSameString(self, span)
        exception = self.start == span.start and self.stop < span.stop
        return self.start < span.start < self.stop or exception
    
    def __ge__(self, span):
        """Returns True if Span is partly on the right of span (the Span object given as parameter). Make sense only for contiguous Span and non-empty ones."""
        _checkSpan(span)
        _checkSameString(self, span)
        exception = self.start == span.start and span.stop < self.stop
        return span.start < self.start < span.stop or exception 

    def union(self, span):
        """
Takes a Span object as entry, and returns a new Span instance, with Span.ranges given by the union of the actual Span.ranges with the span.ranges, when one sees the `ranges` attributes as sets of positions of each instance.

| Parameters | Type | Details |
| --- | --- | --- | 
| `span` | `Span` object | A Span object with same mother string (Span.string) and eventually different ranges that the actual instance. |

| Returns | Type | Details |
| --- | --- | --- | 
| `newSpan` | `Span` object | A `Span` object with `newSpan.ranges` = union of `Span.ranges` and `span.ranges` |

| Raises | Details | 
| --- | --- |
| ValueError | in case the entry is not a Span instance. |
| TypeError | in case the span.string is not the same as Span.string. |
        """
        _checkSpan(span)
        _checkSameString(self,span)        
        newSpan = self.__class__(string=self.string,
                                 subtoksep=self.subtoksep,
                                 ranges=self.ranges).append_ranges(span.ranges)
        return newSpan

    def difference(self, span):
        """
Takes a Span object as entry, and returns a new Span instance with Span.ranges given by the difference of the actual Span.ranges with the span.ranges, when one sees the `ranges` attributes as sets of positions of each instance.

| Parameters | Type | Details |
| --- | --- | --- | 
| `span` | `Span` object | A Span object with same mother string (Span.string) and eventually different ranges that the actual instance. |

| Returns | Type | Details |
| --- | --- | --- | 
| `newSpan` | `Span` object | A `Span` object with `newSpan.ranges` = difference of `Span.ranges` and `span.ranges`. |

| Raises | Details | 
| --- | --- |
| ValueError | in case the entry is not a Span instance. |
| TypeError | in case the span.string is not the same as Span.string. |
        """
        _checkSpan(span)
        _checkSameString(self,span)
        newSpan = self.__class__(string=self.string,
                                 subtoksep=self.subtoksep,
                                 ranges=self.ranges).remove_ranges(span.ranges)
        return newSpan
    
    def intersection(self, span):
        """
Takes a Span object as entry and returns a new Span whose Span.ranges given by the intersection of the actual Span.ranges with the span.ranges, when one sees the `ranges` attributes as sets of positions of each instance..

| Parameters | Type | Details |
| --- | --- | --- | 
| `span` | `Span` object | A Span object with same mother string (Span.string) and eventually different ranges that the actual instance. |

| Returns | Type | Details |
| --- | --- | --- | 
| `newSpan` | `Span` object | A `Span` object with `newSpan.ranges` = intersection of `Span.ranges` and `span.ranges` |

| Raises | Details | 
| --- | --- |
| ValueError | in case the entry is not a Span instance. |
| TypeError | in case the span.string is not the same as Span.string. |
        """
        _checkSpan(span)
        _checkSameString(self,span)
        # one uses the fact that AxB = A+B-(A-B)-(B-A)
        # A-B
        AmB = self.__class__(string=self.string,
                             subtoksep=self.subtoksep,
                             ranges=self.ranges).remove_ranges(span.ranges)
        # B-A
        BmA = self.__class__(string=self.string,
                             subtoksep=self.subtoksep,
                             ranges=span.ranges).remove_ranges(self.ranges)
        # A+B
        newSpan = self.__class__(string=self.string,
                                 subtoksep=self.subtoksep,
                                 ranges=span.ranges).append_ranges(self.ranges)
        # A+B - (A-B)
        newSpan.remove_ranges(AmB.ranges)
        # A+B - (A-B) - (B-A)
        newSpan.remove_ranges(BmA.ranges)
        ranges_ = [r for r in _combineRanges(newSpan.ranges) if r]
        newSpan = self.__class__(string=self.string,
                                 subtoksep=self.subtoksep,
                                 ranges=ranges_)
        # withdraw ranges attribute in case it's an empty list
        if not ranges_:
            newSpan.ranges = ranges_
        return newSpan
    
    def symmetric_difference(self, span):
        """
Takes a Span object as entry, and return a new Span instance whose Span.ranges given by the symmetric difference of the actual Span.ranges with the span.ranges, when one sees the `ranges` attributes as sets of positions of each instance.

| Parameters | Type | Details |
| --- | --- | --- | 
| `span` | `Span` object | A Span object with same mother string (Span.string) and eventually different ranges that the actual instance. |

| Returns | Type | Details |
| --- | --- | --- | 
| `newSpan` | `Span` object | A `Span` object with `newSpan.ranges` = symmetric difference of `Span.ranges` and `span.ranges` |

| Raises | Details | 
| --- | --- |
| ValueError | in case the entry is not a Span instance. |
| TypeError | in case the span.string is not the same as Span.string. |
        """
        _checkSpan(span)
        _checkSameString(self,span)
        # one uses the fact that A/B = (A-B)+(B-A)
        # A-B
        AmB = self.__class__(string=self.string,
                             subtoksep=self.subtoksep,
                             ranges=self.ranges).remove_ranges(span.ranges)
        # B-A
        BmA = self.__class__(string=self.string,
                             subtoksep=self.subtoksep,
                             ranges=span.ranges).remove_ranges(self.ranges)
        # does the union of ranges: (A-B)+(B-A)
        ranges_ = []
        ranges_.extend(AmB.ranges)
        ranges_.extend(BmA.ranges)
        ranges_ = [r for r in _combineRanges(ranges_) if r]
        newSpan = self.__class__(string=self.string,
                                 subtoksep=self.subtoksep,
                                 ranges=ranges_)
        return newSpan
      
    def _prepareSpans(self,ranges,remove_empty):
        """Utility that removes empty ranges and constructs a list of Span objects."""
        if remove_empty:
            ranges = [r for r in ranges if any(r_ for r_ in r) or len(r)==2]
        spans = [self.__class__(string=self.string,
                                ranges=r,
                                subtoksep=self.subtoksep)
                  for r in ranges]
        return spans
    
    def partition(self, start, stop, remove_empty=False):
        """
Split the `Span.string` in three `Span` objects : 
 - `string[:start]`
 - `string[start:stop]`
 - `string[stop:]`
and put all non-empty `Span` objects in a list of `Span` instances.

It acts a bit like the `str.partition(s)` method of the Python
`string` object, but `Span.partition` takes `start` and `stop` 
argument instead of a string. 

| Parameters | Type | Details |
| --- | --- | --- | 
| `start` | int | Starting position of the splitting sequence. |
| `stop` | int | Ending position of the splitting sequence. |
| `remove_empty` | bool. Default is `False` | If `True`, returns a `list of Span` instance with only non-empty `Span` objects. see `__bool__()` method for non-empty `Span` |

| Returns | Type | Details |
| --- | --- | --- | 
| `spans` | `list` of `Span` objects | The `list` object containing the different `Span` objects. |
        """
        start,stop = _startstop(self,start,stop)
        r1,temp_ranges = _cutRanges(self.ranges,start,step=len(self.subtoksep))
        r2,r3 = _cutRanges(temp_ranges,stop-start,step=len(self.subtoksep))
        ranges = [r1,r2,r3]
        spans = self._prepareSpans(ranges,remove_empty)
        return spans
    
    def split(self, cuts, remove_empty=False):
        """
Split a text as many times as there are range entities in the cuts list.
Return a list of `Span` instances.

This is a bit like `str.split(s)` method from Python `string`
object, except one has to feed `Span.split` with a full list
of `range(start,stop)` range objects instead of the string 's' in `str.split(s)`
If the `range(start,stop)` tuples in cuts are given by a regex re.finditer
search on `str(Span)`, the two methods give the same thing. 

| Parameters | Type | Details |
| --- | --- | --- | 
| `cuts` | a list of `range(start,stop,)` range objects. start/stop are integer | Basic usage is to take these cuts from [`re.finditer`](https://docs.python.org/3/library/re.html#re.finditer). The start/end integers are given in the relative coordinate system, that is, in terms of the position in `str(Span)`. |
| `remove_empty` | bool. Default is `False` | If `True`, returns a `list of Span` instance with only non-empty `Span` objects. see `__bool__()` method for non-empty `Span` |

| Return | Type | Details |
| --- | --- | --- | 
| `spans` | `list` of `Span` objects | The `list` object containing the different `Span` objects. |

        """
        _checkRanges(cuts)
        split_ranges = list()
        r1,r_temp = _cutRanges(self.ranges,cuts[0].start,step=len(self.subtoksep))
        r2,r_temp = _cutRanges(r_temp,len(cuts[0]),step=len(self.subtoksep))
        cursor = cuts[0].stop
        split_ranges += [r1,r2,]
        if len(cuts) < 2:
            return self._prepareSpans(split_ranges + [r_temp,], remove_empty)
        for r in cuts[1:]:
            r1,r_temp = _cutRanges(r_temp,r.start-cursor,step=len(self.subtoksep))
            r2,r_temp = _cutRanges(r_temp,len(r),step=len(self.subtoksep))
            split_ranges += [r1,r2,]
            cursor = r.stop
        split_ranges.append(r_temp)
        spans = self._prepareSpans(split_ranges, remove_empty)      
        return spans

    def slice(self, start=0, stop=None, size=1, step=1, remove_empty=False):
        """
Cut the `Span.string` in overlapping sequences of strings of size `size` by `step`,
put all these sequences in separated `Span` objects, and finally 
put all theses objects in a list of `Span` instances.

| Parameters | Type | Details |
| --- | --- | --- | 
| `start` | int | The relative position where to start slicing the Span. |
| `stop` | int | The relative position where to stop slicing the Span. |
| `size` | int | The size of the string in each subsequent Span objects. |
| `step` | int | The number of characters skipped from one Span object to the next one. A character is given by `str(Span)` (called relative position) |

| Returns | Type | Details |
| --- | --- | --- | 
| `spans` | `list` of `Span` objects | The `list` object containing the different `Span` objects. |
        """
        start, stop = _startstop(self,start,stop)
        cuts = [(i,i+size) for i in range(start,stop-size+1,step)]
        slice_ranges = list()
        for start,end in cuts:
            r1,temp_ranges = _cutRanges(self.ranges,start,step=len(self.subtoksep))
            r2,r3 = _cutRanges(temp_ranges,end-start,step=len(self.subtoksep))
            slice_ranges.append(r2)
        spans = self._prepareSpans(slice_ranges,remove_empty)
        return spans



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import setuptools

with open("README.md", "r") as file:
    long_description_ = file.read()
    
setuptools.setup(
    name="tokenspan",
    version="0.7.1",
    author="François Konschelle - IAM CHU Bordeaux France",
    author_email="via.issue@only.please",
    description="Basic tools to tokenize (i.e. to construct atomic-entities/sub-strings of) a string, for Natural Language Processing (NLP). Usefull also for annotation, tree parsing, entity linking, ... (in fact, anything that links a string or its sub-parts to an other object). Key concepts are versatility to other librairies, and freedom to define many concepts on top of a string.",
    long_description=long_description_,
    long_description_content_type="text/markdown",
    license="GNU GENERAL PUBLIC LICENSE v.3",
    url="https://framagit.org/nlp/tokenspan",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Text Processing",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
    ],
    python_requires='>=3.6',
)
